/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

"use strict";

var EXPORTED_SYMBOLS = ["Module"];

var Cu = Components.utils;
var Cc = Components.classes;
var Ci = Components.interfaces;

Cu.import("resource://gre/modules/Services.jsm");
//Cu.import("resource://gre/modules/devtools/Console.jsm");


var Module = {
  _firstCall: false,

  isLoading: function(fsCtx) { // ContentFrameMessageManager
    if (this._firstCall) {
      return false;
    }
    // TODO is this a tab window?
    this._firstCall = true;
    enableModule(fsCtx);
    return true;
  },

  dispose: function(fsCtx) {
    disableModule(fsCtx);
  }
};


function enableModule(fsCtx) {
  Services.obs.addObserver(OnDocumentInserted, "document-element-inserted", false);
  ExtPrefs.start();
}


function disableModule(fsCtx) {
  Services.obs.removeObserver(OnDocumentInserted, "document-element-inserted");
  ExtPrefs.stop();
}


var ExtPrefs = {
  _news: false,
  _social: false,
  _all: false,
  _mode: null,

  start: function() {
    Services.prefs.addObserver("extensions.bbb@multifox.", this, false);
    this._reload();
  },

  stop: function() {
    Services.prefs.removeObserver("extensions.bbb@multifox.", this);
  },


  observe: function(subject, topic, data) {
    this._reload();
  },

  _reload: function() {
    this._news   = this._getBoolPref("news");
    this._social = this._getBoolPref("social");
    this._all    = this._getBoolPref("allsites");
    this._mode   = this._getBlockMode();
  },

  enabledForSocial: function() {
    return this._social;
  },

  enabledForNews: function() {
    return this._news;
  },

  enabledForAll: function() {
    return this._all;
  },

  getBlockMode: function() {
    return this._mode;
  },

  _getBoolPref: function(name) {
    // pref doesn't exist => unchecked => false
    var pref = Services.prefs.getBranch("extensions.bbb@multifox.");
    return pref.getPrefType(name) === pref.PREF_BOOL
         ? pref.getBoolPref(name) : false;
  },

  _getBlockMode: function() {
    var pref = Services.prefs.getBranch("extensions.bbb@multifox.");
    if (pref.getPrefType("blockmode") === pref.PREF_STRING) {
      var val = pref.getCharPref("blockmode");
      switch (val) {
        case "blur":
        case "hide":
          return val;
      }
    }
    return "blur";
  }
};


var L10n = {
  _strings: {},

  getText: function(name) {
    if (name in this._strings) {
      return this._strings[name];
    }
    this._strings[name] = Services.strings.
                            createBundle("chrome://bbb15/locale/strings.properties").
                              GetStringFromName(name);
    return this._strings[name];
  }
};

var m_hosts = {
  // eTLD
  "facebook.com":              "social",
  "twitter.com":               "social",
  "instagram.com":             "social",
  "whatsapp.com":              "social",
  "youtube.com":               "social",
  "disqus.com":                "social",
  "linkedin.com":              "social",
  "foursquare.com":            "social",
  "pinterest.com":             "social",
  "ello.co":                   "social",
  "vine.co":                   "social",
  "em.com.br":                 "news",
  "istoe.com.br":              "news",
  "bonde.com.br":              "news",
  "opovo.com.br":              "news",
  "buzzfeed.com":              "news",
  "estadao.com.br":            "news",
  "clicrbs.com.br":            "news",
  "brasilpost.com.br":         "news",
  "gazetadopovo.com.br":       "news",
  "parana-online.com.br":      "news",
  "cartacapital.com.br":       "news",
  "correiobraziliense.com.br": "news",

  // eTLD + 1
  "plus.google.com":         "social",
  "br.yahoo.com":            "news",
  "www.r7.com":              "news",
  "www.ig.com.br":           "news",
  "www.globo.com":           "news",
  "www.uol.com.br":          "news",
  "www.terra.com.br":        "news",
  "extra.globo.com":         "news",
  "epoca.globo.com":         "news",
  "exame.abril.com.br":      "news",
  "noticias.uol.com.br":     "news",
  "atarde.uol.com.br":       "news",
  "odia.ig.com.br":          "news",
  "noticias.yahoo.com":      "news",
  "veja.abril.com.br":       "news",
  "g1.globo.com":            "news",
  "oglobo.globo.com":        "news",
  "folha.uol.com.br":        "news",
  "ultimosegundo.ig.com.br": "news"
};


var OnDocumentInserted = {
  observe: function(subject, topic, data) {
    var win = subject.defaultView;
    if ((win === null) || (win === undefined)) {
      return; // subject = XMLDocument SVGDocument
    }

    if (!isWeb(win.location.protocol)) {
      return;
    }

    var typeWin = "";
    var tld0 = getETLD(win.location.hostname, 0);
    if (tld0 in m_hosts) {
      typeWin = m_hosts[tld0];
    } else {
      tld0 = getETLD(win.location.hostname, 1);
      if (tld0 in m_hosts) {
        typeWin = m_hosts[tld0];
      }
    }

    var active;
    switch (typeWin) {
      case "news":
        active = ExtPrefs.enabledForNews();
        break;
      case "social":
        active = ExtPrefs.enabledForSocial();
        break;
      default:
        active = ExtPrefs.enabledForAll();
        break;
    }

    if (active) {
      enableWindow(win);
    }
  }
};


function isWeb(protocol) {
  return (protocol === "https:") || (protocol === "http:");
}


function getETLD(host, additionalParts) {
  try {
    return Services.eTLD.getBaseDomainFromHost(host, additionalParts);
  } catch (ex) {
    return host;
  }
}


function enableWindow(win) {
  var obs = new win.MutationObserver(function(mutations) {
    for (var idx1 = 0; idx1 < mutations.length; idx1++) {
      var added = mutations[idx1].addedNodes;
      for (var idx2 = 0; idx2 < added.length; idx2++) {
        var elem = added[idx2];
        if (elem.nodeType === elem.TEXT_NODE) {
          parseTextNode(elem);
        } else if (elem.nodeType === elem.ELEMENT_NODE) {
          parseChildrenNode(elem);
        }
      }
    }
  });

  obs.observe(win.document.documentElement,
              //attributes:true, attributeFilter:["title", "href"]}
              {childList:true, subtree:true, characterData:true});
}


function parseChildrenNode(n) {
  // async children to avoid jank
  n.ownerDocument.defaultView.requestAnimationFrame(function() {
    for (var idx = 0; idx < n.childNodes.length; idx++) {
      var t = n.childNodes[idx];
      if (t.nodeType === t.ELEMENT_NODE) {
        if (t.localName === "a") {
          if (t.href.indexOf("/bbb") > -1) {
            blockNode(n);
            return;
          } else if (t.href.indexOf("/big-brother-brasil") > -1) {
            blockNode(n);
            return;
          }
        }
        parseChildrenNode(t);
      } else if (t.nodeType === t.TEXT_NODE) {
        if (parseTextNode(t)) {
          return;
        }
      }
    }
  });
}


function parseTextNode(node) {
  if (node.parentElement !== null) {
    switch (node.parentElement.localName) {
      case "html":
      case "head":
      case "style":
      case "script":
        return false;
    }
  }
  if (hasBlockedWords(node.data)) {
    blockNode(node);
    return true;
  }
  return false;
}




var m_blocked = {
  "bigfone":   null,
  "estaleca":  null,
  "estalecas": null,
  "BIG":   {"BROTHER": null},
  "big":   {"phone":   null,
            "brother": null},
  "Big":   {"Fone":    null,
            "fone":    null,
            "Brother": null},
  "Brother": {"Brasil": null},
  "bbb":   null,
  "bbbs":  null,

  "BBB":   null,
  "BBBs":  null,

  "bbb9":  null,
  "bbb10": null,
  "bbb11": null,
  "bbb12": null,
  "bbb13": null,
  "bbb14": null,
  "bbb15": null,
  "bbb16": null,
  "bbb17": null,
  "bbb18": null,
  "bbb19": null,
  "bbb20": null,

  "BBB9":  null,
  "BBB10": null,
  "BBB11": null,
  "BBB12": null,
  "BBB13": null,
  "BBB14": null,
  "BBB15": null,
  "BBB16": null,
  "BBB17": null,
  "BBB18": null,
  "BBB19": null,
  "BBB20": null
};

var m_reg = new RegExp("\\W", "g");

function hasBlockedWords(txt) {
  var words = txt.split(m_reg);
  var len = words.length;
  for (var idx = 0; idx < len; idx++) {
    if (words[idx] in m_blocked) {
      var moreWords = m_blocked[words[idx]];
      if (moreWords === null) {
        return true;
      }
      if ((idx + 1) < len) {
        if (words[idx + 1] in moreWords) {
          return true;
        }
      }
    }
  }
  return false;
}


function blockNode(elem) {
  if (elem.nodeType === elem.TEXT_NODE) {
    elem = elem.parentElement;
    if (elem === null) {
      return;
    }
  }

  var host = elem.ownerDocument.defaultView.location.hostname;
  var blocked = null;
  switch (host) {
    case "twitter.com":
      blocked = NodeSiteFinder.twitter(elem);
      break;
    case "instagram.com":
      blocked = NodeSiteFinder.instagram(elem);
      break;
    case "www.facebook.com":
      blocked = NodeSiteFinder.facebook(elem);
      break;
    case "tweetdeck.twitter.com":
      break; // <article>
    case "plus.google.com":
      break; // role=article
  }

  if (blocked === null) {
    blocked = NodeSiteFinder.general(elem);
  }

  if (blocked.nodeType === blocked.TEXT_NODE) {
    blocked = blocked.parentElement;
  }

  if (!findParentTag(blocked, "body")) {
    return; // string outside <body> (e.g. <title>)
  }

  if (findParentBlockedNode(blocked) !== null) {
    return; // node already blocked
  }

  blocked.setAttribute("x-bbb-found", blocked.style.display);
  applyBlock(blocked);
}


function findParentBlockedNode(node) {
  var n = node;
  while (n !== null) {
    if (n.hasAttribute("x-bbb-found")) {
      return n;
    }
    n = n.parentElement;
  }
  return null;
}


function applyBlock(node) {
  // delay layout changes
  node.ownerDocument.defaultView.requestAnimationFrame(function() {
    switch (ExtPrefs.getBlockMode()) {
      case "blur":
        applyBlur(node);
        break;
      case "hide":
        applyHide(node);
        break;
      default:
        throw new Error("");
    }
  });
}


function findParentTag(node, tagName) {
  var n = node;
  while (n !== null) {
    if (n.localName === tagName) {
      return n;
    }
    n = n.parentElement;
  }
  return null;
}


var NodeSiteFinder = {

  general: function(node) {
    var parent = null;
    if (parent === null) {
      parent = findParentTag(node, "li");
    }
    if (parent === null) {
      parent = findParentTag(node, "article");
    }
    if (parent === null) {
      parent = this._findParentAttr(node, "role", "article");
    }

    if (parent === null) {
      var parent1 = findParentTag(node, "div");
      var parent2 = findParentTag(node, "section");
      if ((parent1 !== null) && (parent2 !== null)) {
        // use the inner one
        return parent1.contains(parent2) ? parent2 : parent1;
      } else if ((parent1 !== null) || (parent2 !== null)) {
        return parent1 !== null ? parent1 : parent2;
      } // else { both are null }
    }

    if (parent === null) {
      parent = node;
    }
    return parent;
  },

  // _4-u2 mbm _5jmm _5pat _5v3q
  facebook: function(node) {
    return this._findParentClass(node, "_4-u2");
  },

  instagram: function(node) {
    return this._findParentClass(node, "Comment");
  },


  twitter: function(node) {
    var parent = null;
    if (parent === null) {
      parent = this._findParentClass(node, "js-actionable-tweet");
    }
    if (parent === null) {
      // twitter.com/user/status/12345 (thread) js-simple-tweet
      parent = this._findParentClass(node, "js-simple-tweet");
    }
    return parent;
  },


  _findParentAttr: function(node, attr, val) {
    var n = node;
    while (n !== null) {
      if (n.hasAttribute(attr) && (n.getAttribute(attr) === val)) {
        return n;
      }
      n = n.parentElement;
    }
    return null;
  },


  _findParentClass: function(node, className) {
    var n = node;
    while (n !== null) {
      if (n.classList.contains(className)) {
        return n;
      }
      n = n.parentElement;
    }
    return null;
  }
};


function applyHide(node) {
  var blockId = Math.floor(Math.random() * 8388608);
  node.setAttribute("x-bbb-node-id", blockId);
  var doc = node.ownerDocument;
  var elem = node.parentElement.insertBefore(doc.createElement(node.localName), node);
  elem.setAttribute("x-bbb-label", blockId);
  elem.appendChild(doc.createTextNode(L10n.getText("mode.hide.label")));
  elem.setAttribute("title", L10n.getText("blocked.title"));
  elem.addEventListener("click", function onClick(evt) {
    evt.stopPropagation();
    evt.preventDefault();
    var elem2 = evt.target;
    elem2.removeEventListener("click", onClick);
    var q = "[x-bbb-node-id='" + elem2.getAttribute("x-bbb-label") + "']";
    var node2 = elem2.ownerDocument.querySelector(q);
    if (node2 !== null) {
      node2.style.display = node2.getAttribute("x-bbb-found");
      node2.removeAttribute("x-bbb-found");
    }
    elem2.parentElement.removeChild(elem2);
  });
  elem.style = "color:grey; font-size:11px; cursor:pointer; text-decoration:underline; text-align:center";
  node.style.display = "none";
}


function applyBlur(node) {
  node.setAttribute("title", L10n.getText("blocked.title"));
  node.addEventListener("mouseenter", onMouseEnter);
  node.style.filter = "blur(4px) opacity(50%)";
}


function onMouseEnter(evt) {
  var n2 = evt.target;
  n2.style.filter = "";
  n2.style.transition = "filter .1s ease-in-out";
  n2.addEventListener("mouseleave", function onMouseLeave(evt) {
    var node = evt.target;
    node.removeEventListener("mouseleave", onMouseLeave);
    node.style.transition = "filter .8s ease-in-out";
    node.style.filter = "blur(4px) opacity(50%)";
  });
}
